(define (parse sexp)
  (cond ((or (number? sexp)
             (string? sexp)
             (eqv? sexp #t)
             (eqv? sexp #f)) (make-lit-exp sexp))
  
	  ((and (list? sexp) (eqv? (car sexp) 'quote))
	   (make-lit-exp (cadr sexp)))	       
	       
	  ((symbol? sexp) (make-var-exp sexp))
          
          ((list? sexp)
           (let ((prim (assq (car sexp) *primitives*)))
	     (if prim
		 (make-prim-exp (cadr prim) (parse-exps (cdr sexp)))
                 
                 (case (car sexp)
                   ((if)
                    (make-if-exp (parse (cadr sexp))
                                 (parse (caddr sexp))
                                 (parse (cadddr sexp))))
                   
                   ((env-capture)
                    (make-env-capt-exp))
                   
                   ;;(begin .. .. ..)
                   ((begin)
                    (make-begin-exp (parse-exps (cdr sexp))))
                                      
                   ;; (slet (c d f) ((x 4) (y 3)) ...)
                   ((slet)
                    (make-slet-exp (parse (cadr sexp))
                                   (map car (caddr sexp))
                                   (parse-exps (map cadr (caddr sexp)))
                                   (parse-exps (cdddr sexp))))
                   
                   ;; let and dlet are now just syntactic sugar for scoped let.
                   ((let) ;; standard lexically-scoped let
                    (make-slet-exp (parse '(ds-static))
                                   (map car (cadr sexp))
                                   (parse-exps (map cadr (cadr sexp)))
                                   (parse-exps (cddr sexp))))
                   
                   ((dlet) ;; dynamically-scoped let
                    (make-slet-exp (parse '(ds-dynamic))
                                   (map car (cadr sexp))
                                   (parse-exps (map cadr (cadr sexp)))
                                   (parse-exps (cddr sexp))))
                   
                   ((plet) ;; pervasive let (dynamically-scoped + sticky)
                    (make-slet-exp (parse '(ds-all))
                                   (map car (cadr sexp))
                                   (parse-exps (map cadr (cadr sexp)))
                                   (parse-exps (cddr sexp))))
                   
                   ((flet) ;; flat let (statically-scoped only in direct body (not in lambdas))
                    (make-slet-exp (parse '(ds-flat))
                                   (map car (cadr sexp))
                                   (parse-exps (map cadr (cadr sexp)))
                                   (parse-exps (cddr sexp))))
                   
                   ((soft-let) ;; statically-scoped, but give priority to dynamic bindings
                               ; syntactic limitation: works ONLY with 1 binding (soft-let (x 1) ...)
                    (let ((id (caadr sexp)))
                      (make-slet-exp (parse `(make-ds-soft (ds-static) (quote ,id)))
                                     (list id)
                                     (list (parse (cadr (cadr sexp))))
                                     (parse-exps (cddr sexp)))))
                                     
                   ((soft-plet) ;; pervasive, but give priority to dynamic bindings
                                ; syntactic limitation: works ONLY with 1 binding (soft-plet (x 1) ...)
                    (let ((id (caadr sexp)))
                      (make-slet-exp (parse `(make-ds-soft (ds-all) (quote ,id)))
                                     (list id)
                                     (list (parse (cadr (cadr sexp))))
                                     (parse-exps (cddr sexp)))))
                   
                   ;; (fun (x y z) ...)
                   ((fun)
                    (make-fun-exp (cadr sexp)
                                  (parse-exps (cddr sexp))))
                   
                   ;; (this-jp)
                   ((this-jp) (make-this-jp-exp)) 
                   
                   (else
                    ;; (f 1 2 3)
                    (make-app-exp (parse (car sexp))
                                  (parse-exps (cdr sexp))))
                   ))))))

(define parse-exps (lambda (exps) (map parse exps)))

(define *my-context* #f)

;;; primitives
(define *primitives*
  `((+       ,(lambda args (apply + args)))
    (-       ,(lambda args (apply - args)))
    (*       ,(lambda args (apply * args)))
    (/       ,(lambda args (apply / args)))
    (=       ,(lambda args (apply = args)))
    (<       ,(lambda args (apply < args)))
    (<=      ,(lambda args (apply <= args)))
    (>       ,(lambda args (apply > args)))
    (>=      ,(lambda args (apply >= args)))
    (zero?   ,(lambda args (apply zero? args)))
    (eq?     ,(lambda args (apply eq? args)))

    (not     ,(lambda args (apply not args)))
    (and     ,(lambda (a b) (and a b))) ;;not good! no short-circuit eval
    (write   ,(lambda args (apply write args)))
    (write-string   ,(lambda args (apply write-string args)))
    (printf  ,(lambda args (apply printf args)))
    (newline ,(lambda args (apply newline args)))
    (get-output-string ,(lambda args (apply get-output-string args)))
    (result  ,(lambda () *result*))

    (car     ,(lambda args (apply car  args)))
    (cdr     ,(lambda args (apply cdr  args)))
    (cons    ,(lambda args (apply cons args)))
    (list    ,(lambda args args))
    (length  ,(lambda args (apply length args)))
    (null?   ,(lambda args (apply null? args)))
    
    ;; for defining pointcuts and advices
    (eq-jp?  ,(lambda (jp cl) (eq? (jp-closure jp) cl)))
    (jp-depth ,(lambda (jp) (jp-depth jp)))
 ;   (stack-depth-less  ,(lambda (n jp) (stack-depth-less n jp)))
    (ds-static  ,(lambda () ds-static))
    (ds-dynamic ,(lambda () ds-dynamic))
    (ds-all     ,(lambda () ds-all))
    (ds-flat    ,(lambda () ds-flat))
    (ds-soft     ,(lambda (id) (ds-soft id)))
    (make-ds-soft ,(lambda (ds id) (make-ds-soft ds id)))
    
    (context-set! ,(lambda (val) (set! *my-context* val)))
    (context-get  ,(lambda () *my-context*))
    
    ;; introspecting environments
    (env-check      ,(lambda args (apply env-check args)))
    (env-check-mark ,(lambda args (apply env-check-mark args)))
    (env-find       ,(lambda (id env) (env-lookup id env
                                                  (lambda (pos env) env)
                                                  (lambda (id) #f)
                                                  (lambda (e) #t))))
    (env-mark       ,(lambda (env) (extended-env-mark env)))
    (env-next       ,(lambda (env) (extended-env-env env)))
    (env-print      ,(lambda args (apply print-env args)))
    ))
